'+---------------------------------------------------------------------------+
'| Clusters.bas  --  predicts/reports cluster usages & disk space wasted     |
'|                   based on the cluster sizes used by various operating    |
'|                   systems and how much space is currently in use          |
'|---------------------------------------------------------------------------|
'|Revision History                                                           |
'|---------------------------------------------------------------------------|
'|When        What                                                           |
'| 5 Nov 95   Initial coding and release                                     |
'|19 Nov 95   Fixed problem with counting one too many clusters for          |
'|            each file.  Also fixed the problem with a divide by zero error |
'|            if a drive had only directories (no files) in the root.        |
'|                                                                           |
'|            Added:                                                         |
'|              - the ability to do 512byte clusters (HPFS and NTFS)         |
'|              - Capability to process one drive without zeroing the totals |
'|                                                                           |
'+---------------------------------------------------------------------------+
$STACK 32766
defint a-z
%False = 0
%True = NOT(%False)

%FLAGS = 0                    ' constants for CPU registers -- we will be
%AX    = 1                    ' calling some DOS routines later
%BX    = 2
%CX    = 3
%DX    = 4
%SI    = 5
%DI    = 6
%BP    = 7
%DS    = 8
%ES    = 9

' Define some color constants -- Add sixteen to get blink

%Black   = 0 : %Grey 		= 8
%Blue    = 1 : %BrightBlue  = 9
%Green   = 2 : %BrightGreen = 10
%Cyan    = 3 : %BrightCyan  = 11
%Red     = 4 : %BrightRed   = 12
%Magenta = 5 : %BrightMagenta = 13
%Brown   = 6 : %Yellow		= 14
%White   = 7 : %BrightWhite = 15

TYPE DTAData                  ' all the information we want about a file
	Reserved AS STRING * 21     ' is in this data structure in DOS memory
  Attr     AS BYTE
  FileTime AS WORD
  FileDate AS WORD
  Size     AS DWORD
  TheName  AS STRING * 13
  TheRest  AS STRING * 21
END TYPE

SHARED ReportData&(), TotalFileSize&&, DirNames$(), Pain!, DrvCluster&

%DirMax = 5000

%ClustSizes = 6
DIM ReportData&(1:%ClustSizes,1:5)
'
'  Each primary record in the ReportData& array has 5 elements:
'    ReportData&(x%,1) = Cluster Size
'    ReportData&(x%,2) = Total Clusters Allocated
'    ReportData&(x%,3) = Total Bytes Allocated
'    ReportData&(x%,4) = Total Bytes Used
'    ReportData&(x%,5) = Total Bytes Wasted
'
ReportData&(1,1) = 512        ' initialize the records with default sizes
ReportData&(2,1) = 2048
ReportData&(3,1) = 4096
ReportData&(4,1) = 8192
ReportData&(5,1) = 16384
ReportData&(6,1) = 32768


' Make a title bar
CLS
t$ = "(c) Nathan C. Durland -- ndurland@aol.com"
PRINT sClr$(%BrightWhite,%Blue);"Disk usage prediction program";TAB(81-LEN(t$));t$
PRINT sClr$(%White,%Black);

'+---------------------------------------------------------------------+
'| First, we need to know what the user wants to know                  |
'+---------------------------------------------------------------------+

GetDriveLetter:
VIEW TEXT (1,2)-(80,24)       ' so title bar always stays on screen
CLS
INPUT "Enter drive to project cluster usage for, or QUIT: ";TheDrive$
TheDrive$ = UCASE$(TheDrive$)
IF UCASE$(TheDrive$) = "QUIT" THEN END

REDIM HUGE DirNames$(0:%DirMax)
IF LEN(TheDrive$) = 0 THEN TheDrive$ = CURDIR$
TheDrive$ = RTRIM$(TheDrive$, ANY "\: ") + ":"
DirNames$(0) = LEFT$(UCASE$(TheDrive$),1) + ":\"
CALL GetDriveInfo(TheDrive$, DrvCluster&, ClusterCount&)
IF DrvCluster& < 0 THEN
  PRINT "Could not determine cluster size for drive ";TheDrive$
  END
END IF

'+-----------------------------------------------------------------------+
'| If Pain! is non-zero, then we're not doing the first drive, and need  |
'| to find out what tolerance the user has                               |
'+-----------------------------------------------------------------------+
IF Pain! = 0 THEN
	INPUT "Pain Level: How much (in percent) of slack space is OK";tt$
	Pain! = VAL(tt$)
	IF (Pain! < 1) OR (Pain! > 90) THEN Pain! = 30
END IF

IF ReportData&(1,2) > 0 THEN		                    ' this is not the first run
	DO
  	INPUT "Reset totals to Zero? (Y/N/Q)";tt$
    tt$ = UCASE$(LEFT$(tt$,1))
    IF tt$= "Q" THEN END
  LOOP WHILE INSTR("YN",tt$) = 0

	IF tt$ = "Y" THEN
  	FOR x% = 1 TO %ClustSizes                       ' zero totals
  		ReportData&(x%,2) = 0
  		ReportData&(x%,3) = 0
  		ReportData&(x%,4) = 0
  		ReportData&(x%,5) = 0
  	NEXT x%
  END IF
END IF


'+--------------------------------------------------------------------------+
'| Now we have to fill an array with a list of all the directries on        |
'| the drive -- the array is DIMed to 4000 above, so we should have enough  |
'| capacity for most users hard drives                                      |
'+--------------------------------------------------------------------------+
PRINT "Loading a list of directory names . . ."
CALL ScanForDirs(DirNames$(),DirCount&)

CLS

FileCnt& = 0

'+-------------------------------------------------------------------------+
'| This is the stuff that stays on the screen unchanged.                   |
'+-------------------------------------------------------------------------+

PRINT sClr$(%White,%Black);
PRINT "Processing ";DirCount&;" directories. ";
t$ = USING$("###.##",Pain!)
PRINT "Pain threshold is ";sClr$(%BrightRed,%Black);t$;"%"
PRINT sClr$(%White,%Black);"Drive being processed: ";sClr$(%Yellow,%Black);TheDrive$;
PRINT sClr$(%White,%Black);" Cluster Size: ";sClr$(%Yellow,%Black);DrvCluster&;
PRINT sClr$(%White,%Black);" Total Clusters: ";sClr$(%Yellow,%Black);ClusterCount&

LOCATE 15,1
PRINT sClr$(%Green,%Black);"DOS partion size/cluster size relationship:"
PRINT "      0MB - 128MB  =  2K (2048 byte) Clusters"
PRINT "    129MB - 256MB  =  4K (4096 byte) Clusters"
PRINT "    257MB - 512MB  =  8K (8192 byte) Clusters"
PRINT "    513MB - 1.02GB = 16K (16384 byte) Clusters"
PRINT "   1.02GB - 2.04GB = 32K (32768 byte) Clusters"
PRINT " 512 byte clusters are used by OS/2 (HPFS) & Windows NT (NTFS)";sClr$(%White,%Black)

'+---------------------------------------------------------------------------+
'|  Here, we look at the files in each directory.  We obtain the size of the |
'|  file from DOS, and compute how many clusters that file will consume with |
'|  each of our cluster sizes.  The screen is then updated accordingly       |
'+---------------------------------------------------------------------------+

FOR DirPtr& = 0 TO DirCount&

  LOCATE 3,1                                    ' update the user on what's
  PRINT sClr$(%White,%Black);"Directory ";                ' going on
  t$ = LEFT$(DirNames$(DirPtr&)+SPACE$(75),68)
  PRINT sClr$(%BrightGreen,%Black);t$

  IF INKEY$ = CHR$(27) THEN END                 ' ESC wil abort the program

  FileSpec$ = DirNames$(DirPtr&) + "*.*"
'+--------------------------------------------------------------------------+
'| We use the attribute of 55 so that we get all files - hidden, system,    |
'| archive, -- everything except the volume label                           |
'+--------------------------------------------------------------------------+
  TheFile$ = DIR$(FileSpec$,55)
  WHILE LEN(TheFile$) > 0
    INCR FileCnt&,1
    IF (TheFile$ = ".") OR (TheFile$ = "..") THEN   ' the entries "." and ".."
      TheFile$ = DIR$                               ' represent special dir
      ITERATE                                       ' entries to DOS, and
    END IF                                          ' dont' count

    FileName$ = DirNames$(DirPtr&) + TheFile$       ' An informed user is
    LOCATE 4,1                                      ' a happy user
    PRINT sClr$(%White,%Black);"File:";TAB(8);
    PRINT sClr$(%BrightCyan,%Black);TheFile$;"        ";
    PRINT TAB(30);sClr$(%Yellow,%Black);
    PRINT USING "#,###,###";FileCnt&;
    PRINT sClr$(%White,%Black);" Files and ";sClr$(%Yellow,%Black);
    PRINT USING "##,###";DirPtr&;
    PRINT sClr$(%White,%Black);" Directories checked"

    fSize& = GetFileSize&(FileName$)
    FOR x% = 1 TO %ClustSizes                       ' Compute file stats for
    	CALL ComputeClusters(x%, fSize&)              ' each cluster size
		NEXT x%

    CALL PrintStats                                 ' tell what we found

    TheFile$ = DIR$                                 ' get the next file

  WEND

NEXT DirPtr&

PRINT sClr$(%White,%Black);

DO
	LOCATE 22,1
	INPUT "Would you like to examine another drive? (Y/N)";tt$
  tt$ = UCASE$(LEFT$(tt$,1))
LOOP WHILE INSTR("YN",tt$) = 0

IF tt$ = "Y" THEN GOTO GetDriveLetter

VIEW TEXT (1,1)-(80,25)
locate 24,1

END


SUB ScanForDirs(DirNames$(), DirCount&) LOCAL PUBLIC
'+----------------------------------------------------------------------------+
'|Scans for all the subdirs in a directory tree.                              |
'|	DirPtr& is a pointer to which directory we are currently scanning for     |
'|	more directories.  Each time we find one, we put the name of the new      |
'|	directory on the end of the list, and increase DirCount& by one.  Thus,   |
'|	DirCount& becomes a moving target, and we will just continually move      |
'|	down farther and farther untill all directories have been search to their |
'|  deepest level.                                                            |
'|                                                                            |
'|After we've found all the directories, (DirCount doesn't stay ahead of      |
'|	DirPtr&), we'll sort the list, and exit.                                  |
'|                                                                            |
'|The scan starts at whatever directory is specified in DirNames$(0).  If     |
'| that element is blank, the root directory of the current drive is used     |
'+----------------------------------------------------------------------------+
'
DirPtr& = -1
DirCount& = 0
IF RIGHT$(DirNames$(0),1) <> "\" THEN DirNames$(0) = DirNames$(0) + "\"

WHILE DirCount& > DirPtr&
	INCR DirPtr&,1
	TheDir$ = DirNames$(DirPtr&)
  PRINT "-----> ";TheDir$
'+----------------------------------------------------------------------------+
'| Use an attribute value of 23.  This will make DIR$ return directory names  |
'| and "normal" files, as well has hidden and system directories              |
'+----------------------------------------------------------------------------+
	x$ = DIR$(TheDir$+"*.*",23)
	WHILE x$ <> ""
		TheFile$ = TheDir$ + x$
'+----------------------------------------------------------------------------+
'| Check the attribute of the returned file name.  If bit 4 is set, then the  |
'| file is a directory.  If it's a directory, then increase the DirCount& and |
'| store this directory name.                                                 |
'+----------------------------------------------------------------------------+
    z? = ATTRIB(TheFile$)
		IF BIT(z?,4) THEN
			INCR DirCount&,1
			IF RIGHT$(TheFile$,1) <> "\" THEN TheFile$ = TheFile$ + "\"
			DirNames$(DirCount&) = TheFile$
		END IF
		x$ = DIR$
	WEND
WEND

ARRAY SORT DirNames$() FOR DirCount&+1	' sort directory names

END SUB



SUB ComputeClusters(BYVAL What%, BYVAL k&) SHARED PUBLIC
'+----------------------------------------------------------------------+
'| This little routine computes the cluster statistics for a single     |
'| record in the ReportData& array.                                     |
'|                                                                      |
'|      What%  =  Which Record to compute                               |
'|         k&  =  The size of the current file to be added in           |
'+----------------------------------------------------------------------+

  Cluster& = ReportData&(What%,1)
	TotalFSize&& = TotalFSize&& + k&

  ClustAloc& = (k& \ Cluster&)              ' get number of whole clust
  f& = k& MOD Cluster&                      ' adjust if there is a
  IF f& > 0 THEN INCR ClustAloc&,1          ' partial cluster also
  fWaste& = (ClustAloc& * Cluster&) - k&

  ReportData&(What%,2) = ReportData&(What%,2) + ClustAloc&
  ReportData&(What%,3) = ReportData&(What%,3) + (ClustAloc& * Cluster&)
  ReportData&(What%,4) = ReportData&(What%,4) + k&
  ReportData&(What%,5) = ReportData&(What%,5) + fWaste&

END SUB

SUB PrintStats SHARED PUBLIC
'+---------------------------------------------------------------------------+
'| This routine simply puts the numbers on the screen.  Percentages that are |
'| over the pain threshold will blink                                        |
'+---------------------------------------------------------------------------+

	LOCATE 6,1
  PRINT sClr(%Yellow,%Black);"Clust Size   Clust Alloc    Bytes Alloc    Bytes used        Wasted   %Slack"
  PRINT                      "~~~~~~~~~~   ~~~~~~~~~~~    ~~~~~~~~~~~   ~~~~~~~~~~~       ~~~~~~~  ~~~~~~~"

  FOR x% = 1 TO %ClustSizes
  	PRINT sClr$(%BrightMagenta,%Black);
    IF ReportData&(x%,1) = DrvCluster& THEN x$ = sClr$(%BrightWhite,%Blue)
    PRINT USING " ##,###";ReportData&(x%,1);
    PRINT sClr$(%BrightMagenta,%Black);"        ";
    PRINT USING " ###,###";ReportData&(x%,2);

  	FOR y% = 3 TO 5
    	PRINT USING "  #,###,###,###";ReportData&(x%,y%);
    NEXT y%

    IF ReportData&(x%,3) = 0 THEN ITERATE FOR
    a! = 100 - CFIX((ReportData&(x%,4) / ReportData&(x%,3)) * 100)

    PRINT "  ";
    IF a! < Pain! THEN
    	PRINT sClr$(%BrightWhite,%Black);
    ELSE
    	PRINT sClr$(%Yellow+16,%Black);
    END IF
    PRINT USING " ###.##"; a!

  NEXT x%

END SUB

FUNCTION sClr$(f%,b%) LOCAL PUBLIC
'
'  I use this simply to set the text color in a print statement
'

	COLOR f%,b%
  sClr$ = ""

END FUNCTION


FUNCTION GetFileSize&(BYVAL TheFile$) LOCAL PUBLIC
'+---------------------------------------------------------------------------+
'| Returns the size (in bytes) of the specified file.  Will preserve the     |
'| current DOS DTA. (see below)                                              |
'+---------------------------------------------------------------------------+

  DIM MyDTA as DtaData
  CALL FindOldDta(OldSeg??, OldOff??)	          ' so we can go back
  a$ = SPACE$(LEN(MyDTA))
  NewSeg?? = BITS??(STRSEG(a$))
  NewOff?? = BITS??(STRPTR(a$))
  CALL SetNewDta(NewSeg??, NewOff??)

  REG %AX, &H4E00
  Victim$ = TheFile$ + CHR$(0)
  REG %CX, &H16		' find all files and sub dirs
  REG %DS, STRSEG(Victim$)
  REG %DX, STRPTR(Victim$)
  CALL INTERRUPT &H21

  RetCode?? = REG(%Flags)
  IF BIT(RetCode??,0) THEN                      ' error flag set
    GetFileSize& = -1
  ELSE
    LSET MyDta = a$
    GetFileSize& = MyDta.Size
  END IF

' Reset the old DTA
  CALL SetNewDTA(OldSeg??, OldOff??)
END FUNCTION


'+--------------------------------------------------------------------------+
'| The DTA (disk transfer address) is a data structure in DOS memory that   |
'| holds information about the last file DOS was asked to look for.  DOS    |
'| allows a program to determine where the DTA is, and to set it's own area |
'| for the DTA.                                                             |
'|                                                                          |
'| PowerBASIC uses the DTA when you use the DIR$ function.   The first      |
'| time you use it, you specify a file name: a$ = DIR$("*.*").  After that, |
'| you DOS uses information leftover in the DTA to find the next file when  |
'| you use b$ = DIR$.  Everytime you specify a file, even with wildcards,   |
'| the search starts over.                                                  |
'|                                                                          |
'| In this program, that could be a pain -- our main loop is searching      |
'| through all the file in a directory {x$ = DIR$(TheDir$+"*.*",23)} to     |
'| get the names, then CALLing a sub to get the file size.  It's important  |
'| that the SUB not distrub the DTA, or the program would never get past    |
'| the first file in a directory!                                           |
'|                                                                          |
'| So, these two routines allow us to find & preserve the old DTA data,     |
'| while temporarily substituting a different one for doing the size check  |
'+--------------------------------------------------------------------------+


SUB FindOldDTA(TheSeg??, TheOffSet??) LOCAL PUBLIC
'+------------------------------------------------------------------------+
'|Determines where the current DTA is, and returns it's segment/offset    |
'+------------------------------------------------------------------------+
 REG %AX, &H2F00          'Use DOS service INT 21/function 2F, to
 CALL INTERRUPT &H21      'Find where the current DTA is
 TheSeg?? = REG(%ES)       'Segment of location of default DTA
 TheOffset??  = REG(%BX)   'Offset into the segment for default DTA
END SUB


SUB SetNewDTA(BYVAL TheSeg??, BYVAL TheOffSet??) LOCAL PUBLIC
'+------------------------------------------------------------------------+
'|Installs a new DTA address for DOS to use                               |
'+------------------------------------------------------------------------+
 REG %AX, &H1A00              'Use another Dos Service (21/1A) to set the DTA
 REG %DS, TheSeg??             'Load the segment of it's location
 REG %DX, TheOffSet??          '  and the offset
 CALL INTERRUPT &H21          'Call the DOS service
END SUB

SUB GetDriveInfo(BYVAL TheDrive$, DrvCluster&, ClusterCnt&) LOCAL PUBLIC
'+--------------------------------------------------------------------------+
'| This sub will determine some of the geometry of the drive -- the sector  |
'| size, and the number of sectors per cluster.  From that, we can compute  |
'| the cluster size for this drive                                          |
'+--------------------------------------------------------------------------+

  dd$ = UCASE$(TheDrive$)
  drv% = ASC(dd$) - 64                  ' unlike BIOS functions, A is drive 1
  IF (drv% < 1) OR (drv% > 26) THEN     ' exit if bad disk name
    DrvCluster& = -1
    EXIT SUB
  END IF

  REG %AX, &H3600             ' DOS Function call - INT 21H/36
  REG %DX, drv%               ' Drive to check
  CALL INTERRUPT &H21         ' Call DOS Int


  Sectors&     = REG(%AX)        ' Sectors per cluster
  SectorSize&  = REG(%CX)        ' Sector size in bytes
  TClusters&   = REG(%DX) 	     ' Total Clusters on disk
  DrvCluster&  = Sectors& * SectorSize&

  IF TClusters& > 0 THEN
  	ClusterCnt& = TClusters&
  ELSE
  	ClusterCnt&  = 65536 - TClusters&
  END IF

END SUB

